# -*- coding: utf-8 -*-
"""
Created on Mon Jul  1 15:29:46 2013

@author: delande
"""
# Copyright Dominique Delande
# Provided "as is" without any warranty
#
# Computes the quantum dynamics of the periodically kicked rotor
# by iteration of the evolution operator over one period
# The latter is computed using a "split" technique
# where the kick operator is computed in configuration space (where it is diagonal)
# and the free propagation is computed in momentum space (where it is diagonal)
# Passing from configuration to momentum representation is done a Fast Fourier Transform
# For this reason, it is more efficient to use a basis size which is a power of 2
# The initial state is chosen randomly near zero momentum, in a band defined
# by the half_width_initial_state parameter
# Averaging can be performed over the initial state and (optionally) the quasi-momentum
# The expectation value <p**2> is printed at each kick
# Also the (average) density in momentum space is printed at the final time 

from math import *
import time
import socket
import random
import os
import numpy
from scipy import fftpack

def compute_p2(wave_function,tab_momentum):
#  n=wave_function.size
#  p2=0.0  
#  norm=0.0
#  for i in range(n):
#    p2+=((i+quasi_momentum-n/2)*abs(wave_function[i]))**2
#    norm+=abs(wave_function[i])**2
  p2=numpy.sum((abs(wave_function)*tab_momentum)**2)
  norm=numpy.sum(abs(wave_function)**2)
#  norm=1.0
#  print norm,norm2 
  return p2/norm

# for generating reproducible random sequences
# comment out for irreproducible sequences  
random.seed(123)  
dpi=2.0*pi
hbar=2.0
# We will follow the transition along a straight line
# (K=4.0,epsilon=0.1) -> (K=8.0,epsilon=0.5) in the (K,epsilon) plane
# The script is executed for 3 different sets of parameter
# The following parameters for a very localized regime
# Execution time: 9 seconds
#K=4.0
#epsilon=0.1
#basis_size=64
#number_of_initial_states=400
#final_time=100
# The following for the diffusive regime
# Execution time: 16 seconds
# You can check that the momentum distribution is well described by a Gaussian
#K=8.0
#epsilon=0.5
#basis_size=256
#number_of_initial_states=400
#final_time=100
# The following for the critical regime
# Execution time: 11 seconds
# You can check that the momentum distribution is well described by a Airy(|x|) function
#K=5.81
#epsilon=0.281
#basis_size=128
#number_of_initial_states=400
#final_time=100
# Finally, just for fun, this is also at the critical regime, for much longer time (1.e4 instead of 1.e2)
# Note a small shift of the critical point because of finite-time effects
# Execution time: 60 seconds
# Signal quite noisy, cqn be improved by more averaging
#K=5.85
#epsilon=0.285
#basis_size=512
#number_of_initial_states=10
#final_time=10000

# For the following list, execution time is 18 seconds
K_list=[4.0,5.81,8.0]
epsilon_list=[0.1,0.281,0.5]
basis_size_list=[64,128,256]
number_of_initial_states=100
final_time=200
omega_2=2.67220902067
omega_3=2.01719090925
# put half_width_initial_state to 0 for no averaging over initial state
half_width_initial_state=1
fixed_quasi_momentum=False
quasi_momentum_0=0.0
random_kicked_rotor=False
# Clean files
p2_file=open('qpkr_p2_vs_time_quantum.dat','w')
p2_file.close()
density_file=open('qpkr_quantum_momentum_distribution.dat','w')
density_file.close()
for i_list in range(3):
  K=K_list[i_list]
  epsilon=epsilon_list[i_list]
  basis_size=basis_size_list[i_list]
  n=basis_size
  K_over_hbar=K/hbar
  dpi_over_n=dpi/n
  tab_p2=numpy.zeros(final_time+1)
  tab_density=numpy.zeros(n)
  tab_momentum=numpy.zeros(n)
  tab_spatial_phase=numpy.zeros(n)
  tab_local_phase=numpy.zeros(n)
  tab_kick_phase_factor=numpy.zeros(n,dtype=complex)
  wave_function=numpy.zeros(basis_size,dtype=complex)
  tab_free_prop_phase_factor=numpy.zeros(basis_size,dtype=complex)
  for i in range(n):
    tab_spatial_phase[i]=cos(i*dpi_over_n)
    if (epsilon==0.0):
      tab_kick_phase_factor[i]=complex(cos(K_over_hbar*tab_spatial_phase[i]),-sin(K_over_hbar*tab_spatial_phase[i]))
  for i_state in range(number_of_initial_states):
  # prepare initial state
  # choose the quasi-momentum and populate only few momentum states near p=0 with random phases
    if (fixed_quasi_momentum):
      quasi_momentum=quasi_momentum_0
    else:
      quasi_momentum=random.uniform(-0.5,0.5)
      
    phi_2=random.uniform(0.0,dpi)
    phi_3=random.uniform(0.0,dpi)
  #  print 'Now starts initial state',i_state,quasi_momentum  
    for i in range(n):
      tab_momentum[i]=i+quasi_momentum-n/2 
      if (random_kicked_rotor):
  # This is for the random kicked rotor      
        phase=random.uniform(0.0,dpi)
      else:  
  # and this is for the regular kicked rotor      
        phase=0.5*hbar*(i+quasi_momentum-n/2)**2
      tab_free_prop_phase_factor[i]=complex(cos(phase),-sin(phase))
    wave_function[0:n]=0.0
    for i in range(n/2-half_width_initial_state,n/2+half_width_initial_state+1):
      phase=random.uniform(0,dpi)
      wave_function[i]=complex(cos(phase),sin(phase))/sqrt(2*half_width_initial_state+1)
  #  tab_p2[0]+=compute_p2(wave_function,tab_momentum)
    tab_p2[0]+=numpy.sum((numpy.real(wave_function)*tab_momentum)**2+(numpy.imag(wave_function)*tab_momentum)**2)
  # now start time propagation
    for my_time in range(1,final_time+1):  
  # first free propagation
      wave_function*=tab_free_prop_phase_factor
  # then apply the kick
  # this requires first to convert to configuration space by a FFT
      wave_function=fftpack.ifft(wave_function,overwrite_x=True)
  # then multiply by the kick operator
      if (epsilon!=0.0):
        local_K_over_hbar=K_over_hbar*(1.0+epsilon*cos(omega_2*my_time+phi_2)*cos(omega_3*my_time+phi_3))
        tab_local_phase=local_K_over_hbar*tab_spatial_phase
        tab_kick_phase_factor=numpy.cos(tab_local_phase)-1j*numpy.sin(tab_local_phase)
      wave_function*=tab_kick_phase_factor   
  # back in momentum space by a FFT      
      wave_function=fftpack.fft(wave_function,overwrite_x=True)
  #    tab_p2[my_time]+=compute_p2(wave_function,tab_momentum)
      tab_p2[my_time]+=numpy.sum((numpy.real(wave_function)*tab_momentum)**2+(numpy.imag(wave_function)*tab_momentum)**2)
  # save the final (squared) wavefunction 
    tab_density+=numpy.real(wave_function)**2+numpy.imag(wave_function)**2
  # output the final result
  p2_file=open('qpkr_p2_vs_time_quantum.dat','a')
  print >>p2_file,'# Data generated on',time.asctime(),' using Python script',os.path.abspath( __file__ ),' on computer:',socket.gethostname()
  if (random_kicked_rotor):   
    print >>p2_file,'# Random quasi-periodic kicked rotor'
  else:
    print >>p2_file,'# Regular quasi-periodic kicked rotor'
  if (fixed_quasi_momentum):
    print >>p2_file,'# Fixed quasi-momentum =',quasi_momentum_0
  else:
    print >>p2_file,'# With averaging over quasi-momentum'     
  print >>p2_file,'# K =',K
  print >>p2_file,'# epsilon =',epsilon
  print >>p2_file,'# omega_2 =',omega_2
  print >>p2_file,'# omega_3 =',omega_3
  print >>p2_file,'# With averaging over phi_2 and phi_3'
  print >>p2_file,'# number_of_initial_states =',number_of_initial_states
  print >>p2_file,'# half_width_initial_state =',half_width_initial_state
  print >>p2_file,'# final_time =',final_time
  print >>p2_file,'# basis_size =',n
  for my_time in range(10,final_time+1):  
    print >>p2_file,my_time,tab_p2[my_time]/number_of_initial_states
  print >>p2_file,' '  
  p2_file.close()
  density_file=open('qpkr_quantum_momentum_distribution.dat','a')
  print >>density_file,'# Data generated on',time.asctime(),' using Python script',os.path.abspath( __file__ ),' on computer:',socket.gethostname()
  if (random_kicked_rotor):   
    print >>density_file,'# Random quasi-periodic kicked rotor'
  else:
    print >>density_file,'# Regular quasi-periodic kicked rotor'
  if (fixed_quasi_momentum):
    print >>density_file,'# Fixed quasi-momentum =',quasi_momentum_0
  else:
    print >>density_file,'# With averaging over quasi-momentum'     
  print >>density_file,'# Several trajectories of the standard map, folded in the [0,2*pi]x[0,2*pi] square'
  print >>density_file,'# K =',K
  print >>density_file,'# epsilon =',epsilon
  print >>density_file,'# omega_2 =',omega_2
  print >>density_file,'# omega_3 =',omega_3
  print >>density_file,'# With averaging over phi_2 and phi_3'
  print >>density_file,'# number_of_initial_states =',number_of_initial_states
  print >>density_file,'# half_width_initial_state =',half_width_initial_state
  print >>density_file,'# final_time =',final_time
  print >>density_file,'# basis_size =',n
  for i in range(n):  
    print >>density_file,i-n/2,tab_density[i]/number_of_initial_states
  print >>density_file,' '  
  density_file.close()
  print 'Done K=',K
